public class Auto
{
     int[] PIX;
     int ppm=100;
     double x=0.0,y=0.0,w=0.4,h=0.2,phi=0.0;
     int R=0,G=0,B=255;
     
     int resw=80,resh=60; //Resolution width and height, camera.
     
     int[][][] video = new int[resh][resw][3];
     
     //Speed in meter per seconds, right and left wheel:
     double vL=0.0,vR=0.0;
     
     //actual Speed and angular velocity of center of mass
     double vx=0.0,vy=0.0,omega=0.0;
     
     /**Eckpunkte /Cornerpoints:
                vL
        p5      p4      p3            
        |-------|-------|             ^
        |       *[x,y]  |    ^        | h
        |_______|_______|    | PHI    v
        p0      p1      p2
                vR
                
        <------ w ------>        
     */
     double[][] p = new double[6][2];
     
     public void setPixelPerMeter(int ppm)
     {
         this.ppm=ppm;
     }
     
     public Auto()
     {
          updateCornerpoints();
     }
     public Auto(double x,double y,double w,double h, double phi, int R, int G, int B)
     {
          init(x,y,w,h,phi,R,G,B);
          updateCornerpoints();
     }
     
     public void init(double x,double y,double w,double h, double phi, int R, int G, int B)
     {
         this.x=x;
         this.y=y;
         this.w=w;
         this.h=h;
         this.phi=phi;
         this.R=R;
         this.G=G;
         this.B=B;
     }
 
     public void setSpeed(double vL, double vR)
     {
         this.vL=vL;
         this.vR=vR;
     }

     public void updateScene(int index, Auto[] auto, int[] PIX,float xoffpx, float yoffpx)
     {
          if(auto==null || PIX==null) return;
       
          if(this.PIX==null || this.PIX.length!=PIX.length)
          {
               this.PIX = new int[PIX.length];
          }
          for(int i=0;i<PIX.length;i++)
          {
               this.PIX[i] = PIX[i];
          }
          
          for(int i=0;i<auto.length;i++)
          {
               if(i!=index) //include all other cars in the own scene
               {
                    auto[i].addCarToScene(this.PIX,xoffpx,yoffpx);
               }
          }
     }

     public void addCarToScene(int[] pixels,float xoffpx, float yoffpx)
     {
          double videowidth = Math.sqrt((p[2][0]-p[3][0])*(p[2][0]-p[3][0])+(p[2][1]-p[3][1])*(p[2][1]-p[3][1]));
          double videoheight = Math.sqrt((p[1][0]-p[2][0])*(p[1][0]-p[2][0])+(p[1][1]-p[2][1])*(p[1][1]-p[2][1]));
       
          double pixwidth  = videowidth/(double)resw;
          double pixheight = videoheight/(double)resh;
       
          double ex_x = (p[2][0]-p[3][0])/videowidth; 
          double ex_y = (p[2][1]-p[3][1])/videowidth; 
       
          double ey_x = (p[4][0]-p[3][0])/videoheight; 
          double ey_y = (p[4][1]-p[3][1])/videoheight; 
       
          double startx = p[4][0] + ex_x*pixwidth*0.5 + ey_x*pixheight*0.5; //p4!!!
          double starty = p[4][1] + ex_y*pixwidth*0.5 + ey_y*pixheight*0.5;
       
          for(int i=0;i<video.length;i++)
          {
               for(int k=0;k<video[i].length;k++)
               {
                     //pixel position of video pixel in the world coordinate system
                     double xx = startx + ex_x*pixwidth*(double)k + ey_x*pixheight*(double)i;
                     double yy = starty + ex_y*pixwidth*(double)k + ey_y*pixheight*(double)i;
                     int px = (int)(p2x(xx,xoffpx)+0.5f);
                     int py = (int)(p2y(yy,yoffpx)+0.5f);
                     //video[i][k][0] = (int)red(pixels[px+py*width]);
                     //video[i][k][1] = (int)green(pixels[px+py*width]);
                     //video[i][k][2] = (int)blue(pixels[px+py*width]);
                     pixels[px+py*width] = (R*256+G)*256+B;
               }
          }
     }

     public void updatePosition(double dt)
     {
         omega = 0.5*(vR-vL)/(0.5*h);
         
         double v = vR - omega*(0.5*h);
         double vlength = Math.sqrt((p[2][0]-p[1][0])*(p[2][0]-p[1][0])+(p[2][1]-p[1][1])*(p[2][1]-p[1][1]));
       
              
         double ex_x = (p[2][0]-p[1][0])/vlength; 
         double ex_y = (p[2][1]-p[1][1])/vlength; 
         
         vx = ex_x*v;
         vy = ex_y*v;
         
         x += vx*dt;
         y += vy*dt;
         phi += omega*dt;
     }
 
     /**
     
          | cos  -sin |   
      T = |           |
          | sin   cos |
     
     */
     public void updateCornerpoints()
     {
          p[0][0] = x+(-w/2.0)*Math.cos(phi) - (-h/2.0)*Math.sin(phi);           
          p[0][1] = y+(-w/2.0)*Math.sin(phi) + (-h/2.0)*Math.cos(phi);           
          p[1][0] = x+0.0                    - (-h/2.0)*Math.sin(phi);           
          p[1][1] = y+0.0                    + (-h/2.0)*Math.cos(phi);           
          p[2][0] = x+(w/2.0)*Math.cos(phi)  - (-h/2.0)*Math.sin(phi);           
          p[2][1] = y+(w/2.0)*Math.sin(phi)  + (-h/2.0)*Math.cos(phi);           
          p[3][0] = x+(w/2.0)*Math.cos(phi)  - (h/2.0)*Math.sin(phi);           
          p[3][1] = y+(w/2.0)*Math.sin(phi)  + (h/2.0)*Math.cos(phi);           
          p[4][0] = x+0.0                    - (h/2.0)*Math.sin(phi);           
          p[4][1] = y+0.0                    + (h/2.0)*Math.cos(phi);           
          p[5][0] = x+(-w/2.0)*Math.cos(phi) - (h/2.0)*Math.sin(phi);           
          p[5][1] = y+(-w/2.0)*Math.sin(phi) + (h/2.0)*Math.cos(phi);           
     }
     
     float p2x(double[] p,float xoffpx, float yoffpx)
     {
         return (float)(p[0]+(double)xoffpx/(double)ppm)*(float)ppm;
     }
     float p2y(double[] p,float xoffpx, float yoffpx)
     {
         return (float)(((double)yoffpx/(double)ppm)-p[1])*(float)ppm;
     }
     float p2x(double x,float xoffpx)
     {
         return (float)(x+(double)xoffpx/(double)ppm)*(float)ppm;
     }
     float p2y(double y,float yoffpx)
     {
         return (float)(((double)yoffpx/(double)ppm)-y)*(float)ppm;
     }


     /**
           p3----->p2
           |
           |
           |
           v
           p4
     */
     public void updateVideo(float xoffpx, float yoffpx)
     {
          int[] pixels = this.PIX;
          if(pixels==null) return;
          double videowidth = Math.sqrt((p[2][0]-p[3][0])*(p[2][0]-p[3][0])+(p[2][1]-p[3][1])*(p[2][1]-p[3][1]));
          double videoheight = Math.sqrt((p[1][0]-p[2][0])*(p[1][0]-p[2][0])+(p[1][1]-p[2][1])*(p[1][1]-p[2][1]));
       
          double pixwidth  = videowidth/(double)resw;
          double pixheight = videoheight/(double)resh;
       
          double ex_x = (p[2][0]-p[3][0])/videowidth; 
          double ex_y = (p[2][1]-p[3][1])/videowidth; 
       
          double ey_x = (p[4][0]-p[3][0])/videoheight; 
          double ey_y = (p[4][1]-p[3][1])/videoheight; 
       
          double startx = p[3][0] + ex_x*pixwidth*0.5 + ey_x*pixheight*0.5;
          double starty = p[3][1] + ex_y*pixwidth*0.5 + ey_y*pixheight*0.5;
       
          for(int i=0;i<video.length;i++)
          {
               for(int k=0;k<video[i].length;k++)
               {
                     //pixel position of video pixel in the world coordinate system
                     double xx = startx + ex_x*pixwidth*(double)k + ey_x*pixheight*(double)i;
                     double yy = starty + ex_y*pixwidth*(double)k + ey_y*pixheight*(double)i;
                     int px = (int)(p2x(xx,xoffpx)+0.5f);
                     int py = (int)(p2y(yy,yoffpx)+0.5f);
                     //video[i][k][0] = (int)red(pixels[px+py*width]);
                     //video[i][k][1] = (int)green(pixels[px+py*width]);
                     //video[i][k][2] = (int)blue(pixels[px+py*width]);
                     int p = pixels[px+py*width];
                     video[i][k][0] = (p >> 16) & 0xFF;
                     video[i][k][1] = (p >> 8) & 0xFF;
                     video[i][k][2] = p & 0xFF;
               }
          }
     }
     
     public void show(float xoffpx, float yoffpx)
     {
          stroke(R,G,B);
          line(p2x(p[0],xoffpx,yoffpx),p2y(p[0],xoffpx,yoffpx),p2x(p[2],xoffpx,yoffpx),p2y(p[2],xoffpx,yoffpx));         
          line(p2x(p[2],xoffpx,yoffpx),p2y(p[2],xoffpx,yoffpx),p2x(p[3],xoffpx,yoffpx),p2y(p[3],xoffpx,yoffpx));         
          line(p2x(p[3],xoffpx,yoffpx),p2y(p[3],xoffpx,yoffpx),p2x(p[5],xoffpx,yoffpx),p2y(p[5],xoffpx,yoffpx));         
          line(p2x(p[5],xoffpx,yoffpx),p2y(p[5],xoffpx,yoffpx),p2x(p[0],xoffpx,yoffpx),p2y(p[0],xoffpx,yoffpx));         
          line(p2x(p[1],xoffpx,yoffpx),p2y(p[1],xoffpx,yoffpx),p2x(p[4],xoffpx,yoffpx),p2y(p[4],xoffpx,yoffpx));         
          
          double videowidth = Math.sqrt((p[2][0]-p[3][0])*(p[2][0]-p[3][0])+(p[2][1]-p[3][1])*(p[2][1]-p[3][1]));
          double videoheight = Math.sqrt((p[1][0]-p[2][0])*(p[1][0]-p[2][0])+(p[1][1]-p[2][1])*(p[1][1]-p[2][1]));
       
          double pixwidth  = videowidth/(double)resw;
          double pixheight = videoheight/(double)resh;
       
          double ex_x = (p[2][0]-p[3][0])/videowidth; 
          double ex_y = (p[2][1]-p[3][1])/videowidth; 
       
          double ey_x = (p[4][0]-p[3][0])/videoheight; 
          double ey_y = (p[4][1]-p[3][1])/videoheight; 
       
          double startx = p[4][0]; //p4!!!
          double starty = p[4][1];
          double startx2 = p[1][0]; //p4!!!
          double starty2 = p[1][1];

          double dy = videoheight/20.0;
          
          for(int i=0;i<20;i++)
          {
              double x1 = startx + ey_x*dy*(double)i;
              double y1 = starty + ey_y*dy*(double)i;
              double x2 = startx2 + ey_x*dy*(double)i;
              double y2 = starty2 + ey_y*dy*(double)i;
              line(p2x(x1,xoffpx),p2y(y1,yoffpx),p2x(x2,xoffpx),p2y(y2,yoffpx));         
          }
     }
     
     public void showVideo(float xoffpx, float yoffpx, float SIZE)
     {
          noStroke();
          for(int i=0;i<video.length;i++)
          {
               for(int k=0;k<video[i].length;k++)
               {
                    fill(video[i][k][0],video[i][k][1],video[i][k][2]);
                    rect(xoffpx+SIZE*(float)k,yoffpx+SIZE*(float)i,SIZE,SIZE);
               }
          }
          stroke(R,G,B);
          noFill();
          rect(xoffpx,yoffpx,SIZE*(float)video[0].length,SIZE*(float)video.length);
     }
}
